---------------------------------------------------------------------------------
-- contest.ex  CreativePortal.ca,2004
-- version 3.0.0
--
-- mailto:code@creativeportal.ca
-- http://www.creativeportal.ca                          written by Chris Bensler
---------------------------------------------------------------------------------
-- Derek Parnell's christmas contest 
---------------------------------------------------------------------------------
-- version 2
--  fixed problem handling leading and trailing quotes

-- version 3
--  fixed problem with embedded quotes! :<

---------------------------------------------------------------------------------
constant START = time() -- rule #10
---------------------------------------------------------------------------------

---------------------------------------------------------------------------------
-- get the command line argument ( rule #11 )
constant CMDLINE = command_line()
if length(CMDLINE) < 3 then abort(1) end if -- need a file arg
constant FNAME = CMDLINE[3]
---------------------------------------------------------------------------------

---------------------------------------------------------------------------------
-- some general declarations --
constant FALSE = 0, TRUE = 1
constant EOF = -1

---------------------------------------------------------------------------------
-- MAIN DECLARATIONS --

constant MAXLEN = 20 -- max 20 characters, rule #6

integer   wcount      wcount    = 0 -- total word count
integer   ucount      ucount    = 0 -- total unique words
sequence  wfreq       wfreq     = repeat(repeat(0,MAXLEN),2) -- table of top word frequencies
sequence  wlengths    wlengths  = repeat(0,MAXLEN) -- table of word lengths

sequence  wtable
wtable = repeat( -- sort by word length
            repeat( -- sort by first char
              repeat( -- sort by last char
                { -- the buckets
                  {} -- {tally1,tally2,...} (unique count for corresponding word)
                 ,{} -- {word1,word2,...}
                }
              ,#80) -- (#7F + 1) by last char. using overkill table size for exact matching
            ,#80)   -- (#7F + 1) sort by first char.
          ,MAXLEN)  -- by length


sequence  wbuffer     wbuffer     = repeat(0,MAXLEN+1) -- word buffer
integer   wlen        wlen        = 0 -- current word length, and first bucket index
integer   whash2      whash2      = 0 -- second bucket index
integer   whash3      whash3      = 0 -- third bucket index

sequence theword -- the actual current word

integer   alpha_flag  alpha_flag  = FALSE -- flag is TRUE if any chars are alpha
integer   lastquote   lastquote = 0   -- equals wlen if last char in the token is a '

integer fn
integer c -- current character
---------------------------------------------------------------------------------

---------------------------------------------------------------------------------

-------------------------------------------------
procedure skip_word()
  -- skip the current word
  while TRUE do
    if (c < '0') or (c > 'z') then -- not a digit or an alpha character
      if (c != '\'') and (c != '-') then -- not a ' or a -
        return
      end if
    elsif (c > '9') and (c < 'a') then -- not a digit or a lower alpha character
      if (c < 'A') or (c > 'Z') then -- not an UPPER alpha character
        return
      end if
    end if
    c = getc(fn)
  end while
end procedure
-------------------------------------------------

-------------------------------------------------
procedure next_word()
  -- skip to the beginning of the next word

  alpha_flag = FALSE -- reset the alpha flag
  wlen = 1 -- initialize the word length
  while TRUE do
    -- check if it's in the widest allowable range
    if (c >= '0') and (c <= 'z') then
      -- narrow it down
      if (c > '9') then -- not a digit
        if (c < 'a') then -- not a lower alpha character
          if (c >= 'A') and (c <= 'Z') then -- it's an UPPER ALPHA character
            c += ' ' -- convert to lowercase
            alpha_flag = TRUE
            wbuffer[1] = c
            return
          end if
        else -- lower alpha character
          alpha_flag = TRUE
          wbuffer[1] = c
          return
        end if
      else -- digit character
        wbuffer[1] = c
        return
      end if

    -- exception
    elsif (c = '\'') then -- valid word delimiter
      alpha_flag = TRUE
      wlen = 0
      return
    elsif (c = '-') then -- invalid word delimiter, but valid word char
      skip_word()
      wlen = 0
      return
    elsif (c = EOF) then -- end of file reached
      wlen = 0
      return
    end if

    c = getc(fn)
  end while
end procedure
-------------------------------------------------

-------------------------------------------------
procedure scan_word()
  lastquote = FALSE
  c = getc(fn)
  
  -- scan while valid word character
  while TRUE do
    -- check if it's in the widest allowable range
    if (c >= '0') and (c <= 'z') then
      -- narrow it down
      if (c > '9') then -- not a digit
        if (c < 'a') then -- not a lower alpha character
          if (c < 'A') or (c > 'Z') then -- not an UPPER ALPHA character
            return -- word is finished
          else -- UPPER ALPHA character
            c += ' ' -- convert to lowercase
          end if
        end if
        -- an alpha character was found in the word ( rule #3 )
        alpha_flag = TRUE
      end if

      -- increment the word length, and store the character
      wlen += 1
      wbuffer[wlen] = c

    -- exceptional cases
    elsif (c = '\'') then -- ignore ' quotes ( rule #5 )
      alpha_flag = TRUE
      lastquote = wlen + 1
    elsif (c = '-') then -- hyphens are valid word characters
      wlen += 1
      wbuffer[wlen] = c
    else -- invalid word character
      return
    end if

    if (wlen > MAXLEN) then -- the current word is too long
      skip_word() -- skip the rest of the current word
      wlen = 0 -- reset the word len
      return -- abort procedure
    end if

    c = getc(fn)
  end while
end procedure
-------------------------------------------------

---------------------------------------------------------------------------------

---------------------------------------------------------------------------------
-- I would prefer to put the stat routines above the parser routines
-- but it seems the most frequent routines should be processed by eu first,
--  so they are found earlier in Eu's internal lookups.
-- I gain a couple of percent on larger files

constant
      WCOUNT = 1
     ,WTABLE = 2

-------------------------------------------------
 -- speed gained by making these local instead of private
integer last,len,cmp -- used in update_freq()
integer f,count      -- used in update_freq() and store_word()
-- the variables won't need to be redefined for each routine instance

procedure update_freq()
  last = MAXLEN

  -- update the word frequency table
  if (count >= wfreq[WCOUNT][20]) then -- the new word needs to be added or updated in the wfreq table

    -- check if the word is already in the table
    f = find(theword,wfreq[WTABLE])
    if f then -- found it, we just have to update the freq count
      last = f
    end if
    -- insert sorted
    f = 1
    if (count <= wfreq[WCOUNT][1]) then -- we have to search the wfreq table
      while (f <= MAXLEN) do
        if (count >= wfreq[WCOUNT][f]) then
          while (wfreq[WCOUNT][f] = count) do
            -- we have to compare textually ( rule #12.c )
            -- sort ascending
            len = length(wfreq[WTABLE][f])
            if (wlen < len) then len = wlen end if

            cmp = compare(theword,wfreq[WTABLE][f][1..len])
            if cmp < 0 then
              exit
            elsif not cmp then
              if (wlen < length(wfreq[WTABLE][f])) then
                exit
              end if
            end if

            f += 1
            if (f > MAXLEN) then exit end if
          end while
          exit
        end if
        f += 1
      end while
    end if

    if (f <= last) then
      if (f < last) then
        wfreq[WCOUNT][f+1..last] = wfreq[WCOUNT][f..last-1]
        wfreq[WTABLE][f+1..last] = wfreq[WTABLE][f..last-1]
      end if
      wfreq[WCOUNT][f] = count
      wfreq[WTABLE][f] = theword
    end if
  end if
end procedure
-------------------------------------------------

-------------------------------------------------
procedure store_word()

  -- increment the total word count
  wcount += 1
  -- increment the word length frequency
  wlengths[wlen] += 1
  
  -- get the hash indexes for the word
  whash2 = wbuffer[1] + 1 -- offset by 1, because NULL is valid
  whash3 = wbuffer[wlen] + 1 -- offset by 1, because NULL is valid

  -- store the actual word, so we don't have to keep slicing
  theword = wbuffer[1..wlen]

  -- check if the current word already exists
  f = find(theword,wtable[wlen][whash2][whash3][WTABLE])

  if f then -- the word already exists in the table
    -- tally the frequency count for the current word
    wtable[wlen][whash2][whash3][WCOUNT][f] += 1
    count = wtable[wlen][whash2][whash3][WCOUNT][f]
  else -- a new word, add it to the table
    -- increment unique word count
    ucount += 1
  
    wtable[wlen][whash2][whash3][WCOUNT] &= 1
    wtable[wlen][whash2][whash3][WTABLE] &={ theword }

    count = 1
  end if

  update_freq()
end procedure
-------------------------------------------------

---------------------------------------------------------------------------------

-- BEGIN --
---------------------------------------------------------------------------------
fn = open(FNAME,"rb")
---------------------------------------------------------------------------------
c = getc(fn)
while c != EOF do

  next_word()
  
  -- check the word, store it, and update the stats
  if wlen or (c = '\'') then
    scan_word() -- scan the next word
    if wlen and alpha_flag then
      if (wbuffer[wlen] != '-') or (lastquote = wlen) then -- found a valid word
        -- store the word, and update the stats
        store_word()
      end if
    end if
  end if
end while
---------------------------------------------------------------------------------
close(fn)
---------------------------------------------------------------------------------

-- DISPLAY --
---------------------------------------------------------------------------------
puts(1,   "Chris Bensler "&FNAME&"\n")
printf(1, "Total: %7d, Unique: %7d\n", {wcount, ucount})
---------------------------------------------------------------------------------

---------------------------------------------------------------------------------
-- display the most frequent words in descending order
-- descending count, ascending text sort
-- display word length frequencies in ascending order of length
-- floor(ucount/1000)+1 or 20
integer num_freq      num_freq = floor(ucount/1000)+1
if (num_freq > MAXLEN) then
  num_freq = MAXLEN
end if

for i = 1 to num_freq do
  printf(1, "%02d %-20s %d\n" , {i, wfreq[WTABLE][i], wfreq[WCOUNT][i]})
end for
---------------------------------------------------------------------------------

---------------------------------------------------------------------------------
-- determine the longest word
integer longest_word      longest_word = 0
for i = MAXLEN to 1 by -1 do
  if wlengths[i] then
    longest_word = i
    exit
  end if
end for

-- display word length frequencies in ascending order of length
for i = 1 to longest_word do
  printf(1, "%02d %d\n", {i,wlengths[i]})
end for
---------------------------------------------------------------------------------

---------------------------------------------------------------------------------
printf(1, "Elapsed time: %f\n", time() - START) -- rule #10
---------------------------------------------------------------------------------
