-- * Euphoria Coding Contest November 2004
-- Author: Elliott Sales de Andrade
-- File: token.e
-- Description: Tokenizes a file according to rules...

--*--------------------------------------------------------------------*--
--* File-level Variables                                               *--
--*--------------------------------------------------------------------*--

-- Warn about invalid characters in file? (Valid: #00..#7F)
global integer gWarnInvalidChars
gWarnInvalidChars = FALSE

-- file handle
integer fn
-- next character input from file
integer c

--*--------------------------------------------------------------------*--
--* Tokenizing Functions                                               *--
--*--------------------------------------------------------------------*--

-- Test for valid token characters...
--  Rule #2
--  1 Parameter: some character
--  Returns: TRUE if it's a valid token character, FALSE otherwise
function IsTokenChar(integer x)
-- This funny looking if-block is faster than the one-liner...
if x <= 'z' then
	if x >= 'a' then
		-- between 'a' and 'z'
		return TRUE
	elsif x > 'Z' then
		-- between 'Z' and 'a'
		return FALSE
	elsif x >= 'A' then
		-- between 'A' and 'Z'
		return TRUE
	elsif x > '9' then
		-- between '9' and 'A'
		return FALSE
	elsif x >= '0' then
		-- between '0' and '9'
		return TRUE
	elsif x = '\'' then
		-- special case in remaining characters...
		return TRUE
	elsif x = '-' then
		-- same as above...
		return TRUE
	else
		-- all others...
		return FALSE
	end if
else
	-- too big
	return FALSE
end if
end function

-- Get next character from file...
--  No Parameters
--  No Return Value
procedure GetChar()
	c = getc(fn)
	-- Check for invalid characters...
	if c >= -1 and c <= #7F then
		return
	else
		gWarnInvalidChars = TRUE
	end if
end procedure

-- Skip any non-token characters
--  No Parameters
--  No Return Value
procedure SkipWhitespace()
	while not IsTokenChar(c) and c != -1 do
		GetChar()
	end while
end procedure

-- These as file-level variables really do
--  seem to help with speed...
sequence tok	-- the token
integer maxLen	-- maximum length of the token (when we need to resize the buffer)
-- Gets the next token from the input file...
--  No parameters
--  Returns: the next token, or "" if EOF
global function GetNextToken()
integer ind,	-- token indexing variable
		hasLetters,	-- is token composed more than just numbers and dashes? Rule #3a
		len		-- length of the current token
tok = repeat(0, 50)
maxLen = 50
len = 0
SkipWhitespace()
if c = '-' then
	-- Tokens starting with '-' are invalid.... Rule #3b
	--  Might as well get rid of characters without wasting time storing them...
	while IsTokenChar(c) do
		GetChar()
	end while
	return GetNextToken()
end if
while IsTokenChar(c) do
	len += 1
	if len > maxLen then
		maxLen += 25
		tok &= repeat(0, 25)
	end if
	-- convert lower case to upper case for easier use
	if c >= 'a' then
		tok[len] = c - 32
	else
		tok[len] = c
	end if
	GetChar()
end while
if len then
	-- Tokens ending with '-' are invalid... Rule #3b
	if tok[len] = '-' then
		return GetNextToken()
	end if
	ind = 1
	hasLetters = FALSE
	-- loop through characters removing quotes,
	--  testing rule #3b, changing lower case to upper case
	while ind <= len do
		-- characters making it a valid token (opposite of Rule #3b)
		if tok[ind] >= 'A' then
			hasLetters = TRUE
			ind += 1
		-- quotes are ignored for effective length (Rule #7)
		elsif tok[ind] = '\'' then
			tok = tok[1..ind - 1] & tok[ind + 1..len]
			len -= 1
			hasLetters = TRUE
		-- just advance the index
		elsif tok[ind] >= '0' then
			ind += 1
		-- same here...
		elsif tok[ind] = '-' then
			ind += 1
		end if
	end while
	if hasLetters = FALSE or  -- Token was only numbers and/or hyphens (Rule #3a)
		len > 20 or			  -- Token is too long (Rule #3c)
		len = 0 then		  -- Effective length of zero (Rule #9)
		return GetNextToken()
	end if
else
	-- If the length of token is 0,
	--  then we have reached EOF
end if
	return tok[1..len]
end function

--*--------------------------------------------------------------------*--
--* File I/O Functions                                                 *--
--*--------------------------------------------------------------------*--

-- Open a file so it can be accessed from this include only...
--  1 Parameter: Name of file
--  Returns: TRUE is successful, FALSE, otherwise
global function OpenFile(sequence filename)
fn = open(filename, "rb")
if fn != -1 then
	-- prep the tokenizer...
	GetChar()
	return TRUE
end if
	return FALSE
end function

-- Calculates size of input file...
--  No Parameters
--  Returns: Size of file in bytes
global function GetFileSize()
integer size, oldpos
-- don't want to waste time including another file...
oldpos = machine_func(20, fn)		-- where(fn)
size = machine_func(19, {fn, -1})	-- seek(fn, -1)
size = machine_func(20, fn)			-- where(fn)
machine_proc(19, {fn, oldpos})		-- seek(fn, oldpos)
	return size
end function

-- For completeness' sake, close the file handle...
--  No Parameters
--  No Return Value
global procedure CloseFile()
	close(fn)
end procedure
