constant START = time()
----------------------------------------
-- jcount5.ex
-- Jason Gade
-- For Derek Parnell's Euphoria Contest
-- Version 5
-- 22 November 2004
--
-- Euphoria coding contest
-- Count the number of words in a file, where the set of words is constrained by the following
-- criteria:
--    Words consist of upper and lower case letters, digits 0-9, single quote and dash;
--    For the purposes of comparison, case does not matter and quotes are not counted;
--    Words consisting of only digits, or digits and dashes, are not counted as words
--    unless they are quoted;
--    Words of zero length after quotes are removed are not counted;
--    Words beginning or ending with a dash are not counted.
--
--    This is a complete rewrite, but still with cut-and-paste from previous
--    submissions.  This is actually my fourth submission.
--
----------------------------------------



----------------------------------------
-- For testing
-- with profile
-- with profile_time
-- with trace
----------------------------------------



----------------------------------------
-- Comment the following lines for debugging
without type_check
without warning
----------------------------------------



----------------------------------------
-- Some useful constants
----------------------------------------

constant
   TRUE   = 1
  ,FALSE  = 0
  ,EOF = -1
  ,STDIN  = 0
  ,STDOUT = 1
  ,STDERR = 2

constant
   AUTHOR = "Jason Gade"

constant
   MAX_TOKEN_LENGTH = 20
  ,MAX_FREQ_TOKENS = 20
  ,EMPTY_TOKEN = repeat(0, MAX_TOKEN_LENGTH)
  ,EMPTY_WORDLIST = repeat(0, 4096)

-- Fields for dereferencing token string and token count.
constant
   STRING = 1
  ,COUNT = 2



----------------------------------------
-- Check options and open file
----------------------------------------

constant
   OPTS = command_line()

if length(OPTS) != 3 then
   puts(STDERR, "Usage: " & OPTS[2] & " filename\n")
   abort(1)
end if

constant
   FILENAME = OPTS[3]
  ,INFILE = open(FILENAME, "rb") -- Use binary to avoid any number problems

if INFILE = -1 then
   puts(STDERR, "Could not open " & FILENAME & "\n")
   abort(1)
end if



----------------------------------------
-- Program variables
----------------------------------------

integer
   Total_Count
  ,Unique_Count
  ,Num_Tokens_To_Display
  ,duplicate

sequence
   Token_Length_Freq
  ,Tokens_By_Freq



----------------------------------------
-- Includes (inlined)
----------------------------------------


---------------------------------------
-- include tokenlist.e
-- tokenlist.e
-- Hashes tokens and adds them to a table
---------------------------------------

-- Lookup table for pearson hash.  This was precalculated and inserted.  It is
-- so long so I don't have to do any division or remainder.
-- Note: since I'm only using ASCII up to 90, I shouldn't need any index greater
-- than 346, but I'm playing it safe and just doubling the whole table.
constant
   TABLE =
   {154, 111, 256, 131, 134, 110, 162, 53, 214, 199, 98, 138, 239, 96, 251, 165
   ,240, 89, 229, 168, 244, 223, 80, 21, 28, 17, 143, 235, 186, 94, 189, 64
   ,178, 141, 61, 97, 158, 133, 222, 46, 173, 42, 232, 197, 65, 37, 102, 114
   ,81, 45, 92, 108, 212, 63, 129, 137, 14, 149, 33, 188, 215, 152, 231, 90
   ,71, 77, 54, 70, 179, 198, 194, 12, 156, 245, 4, 1, 250, 122, 83, 193
   ,253, 217, 43, 25, 220, 115, 3, 30, 171, 58, 218, 5, 2, 128, 203, 205
   ,216, 227, 206, 224, 41, 67, 196, 124, 213, 82, 27, 182, 200, 120, 159, 87
   ,35, 44, 36, 47, 26, 254, 146, 118, 157, 99, 69, 153, 150, 39, 101, 226
   ,105, 161, 249, 104, 132, 167, 183, 187, 31, 113, 163, 93, 181, 175, 145, 139
   ,242, 228, 230, 192, 117, 59, 211, 16, 86, 109, 241, 172, 56, 176, 40, 75
   ,177, 88, 15, 123, 126, 48, 201, 23, 57, 136, 91, 233, 243, 160, 207, 95
   ,155, 60, 100, 29, 22, 68, 148, 208, 237, 119, 236, 51, 164, 8, 127, 112
   ,103, 204, 32, 140, 9, 191, 190, 79, 195, 184, 116, 55, 210, 142, 169, 247
   ,151, 174, 147, 19, 221, 62, 135, 248, 246, 130, 107, 74, 219, 234, 76, 38
   ,50, 20, 78, 85, 185, 255, 49, 72, 11, 7, 238, 66, 225, 52, 24, 121
   ,34, 84, 170, 13, 125, 18, 144, 10, 202, 6, 166, 106, 180, 73, 252, 209
   ,154, 111, 256, 131, 134, 110, 162, 53, 214, 199, 98, 138, 239, 96, 251, 165
   ,240, 89, 229, 168, 244, 223, 80, 21, 28, 17, 143, 235, 186, 94, 189, 64
   ,178, 141, 61, 97, 158, 133, 222, 46, 173, 42, 232, 197, 65, 37, 102, 114
   ,81, 45, 92, 108, 212, 63, 129, 137, 14, 149, 33, 188, 215, 152, 231, 90
   ,71, 77, 54, 70, 179, 198, 194, 12, 156, 245, 4, 1, 250, 122, 83, 193
   ,253, 217, 43, 25, 220, 115, 3, 30, 171, 58, 218, 5, 2, 128, 203, 205
   ,216, 227, 206, 224, 41, 67, 196, 124, 213, 82, 27, 182, 200, 120, 159, 87
   ,35, 44, 36, 47, 26, 254, 146, 118, 157, 99, 69, 153, 150, 39, 101, 226
   ,105, 161, 249, 104, 132, 167, 183, 187, 31, 113, 163, 93, 181, 175, 145, 139
   ,242, 228, 230, 192, 117, 59, 211, 16, 86, 109, 241, 172, 56, 176, 40, 75
   ,177, 88, 15, 123, 126, 48, 201, 23, 57, 136, 91, 233, 243, 160, 207, 95
   ,155, 60, 100, 29, 22, 68, 148, 208, 237, 119, 236, 51, 164, 8, 127, 112
   ,103, 204, 32, 140, 9, 191, 190, 79, 195, 184, 116, 55, 210, 142, 169, 247
   ,151, 174, 147, 19, 221, 62, 135, 248, 246, 130, 107, 74, 219, 234, 76, 38
   ,50, 20, 78, 85, 185, 255, 49, 72, 11, 7, 238, 66, 225, 52, 24, 121}

integer
   len
  ,hval1
  ,hval2

-- Sets hval1 and hval2 to a number from 1-256 representing string
-- Adapted from the Pearson hash by Bob Jenkins http://burtleburtle.net/bob/hash/pearson.html
procedure Hash(sequence word)

   len = length(word)
   hval1 = len
   hval2 = len + 127

   -- Mix it up
   for i = 1 to len do
      hval1 = TABLE[hval1] + word[i]
      hval2 = TABLE[hval2] + word[i]
   end for

   -- Keep the values in range
   hval1 = TABLE[hval1]
   hval2 = TABLE[hval2]

end procedure

sequence
   word
  ,wordList
  ,hashTable

wordList = EMPTY_WORDLIST
hashTable = repeat(repeat({}, 256), 256)

sequence
   bucket

integer
   index

-- Get a word from the hash table.
-- Sets index to point to the word's position in wordList, else sets it to zero.
procedure GetHashWord(sequence string)

  index = 0
  Hash(string)
  bucket = hashTable[hval1][hval2]

  if length(bucket) = 0 then return end if -- empty bucket
  -- search this bucket for the string
  for i = 1 to length(bucket) do
    word = wordList[bucket[i]][STRING]
    if equal(string, word) then -- Found it!
       index = bucket[i]
       exit
    end if
  end for

end procedure

-- Add a string to the hash table
-- If string is not in the table already, add it, with a count of 1,
-- otherwise just increment its count.
procedure AddTokenList(sequence string)

  GetHashWord(string)
  if index then -- Word was found so increment its count
    wordList[index][COUNT] += 1
  else  -- Word wasn't found so add new string with count 1:
    Unique_Count += 1
    index = Unique_Count
    if index > length(wordList) then
       wordList &= EMPTY_WORDLIST
    end if
    wordList[index] = {string, 1}
    hashTable[hval1][hval2] &= {index}
  end if

end procedure



---------------------------------------
-- Set up a table for valid characters and to translate from upper to lower case
----------------------------------------

sequence
   translate
   translate = repeat(0, 256)

-- All the '+ 1's are because of 1-base indexing.  ASCII zero translates to
-- translate[1].
translate['\'' + 1] = '\''
translate['-'+ 1] = '-'
translate['0' + 1..'9' + 1] = "0123456789"
translate['A' + 1..'Z' + 1] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
translate['a' + 1..'z' + 1] = translate['A' + 1..'Z' + 1]



----------------------------------------
-- Main loop:
-- Get words from file and add them to collection
----------------------------------------

integer
   c
  ,char
  ,last
  ,quoted
  ,alpha
  ,tokenLength

object
   token

Unique_Count = 0
Total_Count = 0
Token_Length_Freq = repeat(0, MAX_TOKEN_LENGTH)
Tokens_By_Freq = {}
token = EMPTY_TOKEN
tokenLength = length(token)

-- Get Tokens

  -- Loop through file
  c=TRUE
  while c do -- C can be EOF

  last = 0
  quoted = 0
trace(1)
    -- Loop until delimiter is reached or end of file
    while TRUE do
       c = getc(INFILE)+1
       if c then -- not EOF
          if translate[c] then
             last += 1
             if last > tokenLength then -- make more room
                token &= EMPTY_TOKEN
                tokenLength += MAX_TOKEN_LENGTH
             end if
             token[last] = translate[c]
             if c = '\''+1 then quoted += 1 end if
          elsif last then
             if last - quoted > MAX_TOKEN_LENGTH then
                last = 0
                quoted = 0
             else exit end if -- Too long
          end if
       else exit -- EOF
       end if
    end while

    -- Validate token
    alpha = FALSE

    if last then -- Check for zero-length token
       -- tokens cannot begin or end with a dash reqardless of quotes
       if token[last] != '-' then
          if token[1] != '-' then

             -- Find and strip quotes
             if quoted then
                char = last
                while char do
                   if token[char] = '\'' then
                      token = token[1..char-1] &
                              token[char+1..last] &
                              token [last+1..tokenLength] & 0
                      last -= 1
                   end if
                   char -= 1
                end while
             end if

             -- Check for alpha
             if token[1] < 'A' then
                for i = 2 to last do
                   if token[i] >= 'A' then
                      alpha = TRUE
                      exit
                   end if
                end for
             else
                alpha = TRUE
             end if

             -- Do final validation and add a good token to the list
             if alpha or quoted then
                if last then
                   Total_Count += 1
                   Token_Length_Freq[last] += 1
                   AddTokenList(token[1..last])
                end if
             end if -- alpha or quoted

          end if -- token doesn't end with a '-'

       end if -- token doesn't start with a '-'

    end if -- last > 0

  end while



----------------------------------------
-- Create a list of tokens sorted by most frequent
----------------------------------------

integer
   minCount
  ,tempCount
  ,tokenCount

Num_Tokens_To_Display = floor(Unique_Count/1000)+1

if Num_Tokens_To_Display > MAX_FREQ_TOKENS then
   Num_Tokens_To_Display = MAX_FREQ_TOKENS
end if

minCount = 0

Tokens_By_Freq = repeat({0, 0}, Num_Tokens_To_Display)

for i = 1 to Unique_Count do
   tempCount = wordList[i][COUNT]

   if tempCount >= minCount then -- This goes in the list

      for j = 1 to Num_Tokens_To_Display do
         tokenCount = Tokens_By_Freq[j][COUNT]

         if (tempCount > tokenCount) or
           ((tempCount = tokenCount) and
             compare(wordList[i][STRING], Tokens_By_Freq[j][STRING]) < 0) then -- Insert

            Tokens_By_Freq = Tokens_By_Freq[1..j-1] &
                             {wordList[i]} &
                             Tokens_By_Freq[j..Num_Tokens_To_Display-1]

            minCount = Tokens_By_Freq[Num_Tokens_To_Display][COUNT]
            exit
         end if

      end for

   end if

end for




----------------------------------------
-- Print Results
----------------------------------------

printf(STDOUT, "%s %s\n", {AUTHOR, FILENAME})
printf(STDOUT, "Total: %7d, Unique: %7d\n", {Total_Count, Unique_Count})

for Rank = 1 to Num_Tokens_To_Display do
   token = Tokens_By_Freq[Rank]
   printf(STDOUT, "%02d %-20s %d\n" , {Rank, token[STRING], token[COUNT]})
end for

duplicate = 0
for Token_Length = 1 to MAX_TOKEN_LENGTH do
   token = Token_Length_Freq[Token_Length]
   if token and token != duplicate then
      printf(STDOUT, "%02d %d\n", {Token_Length, token})
   end if
   duplicate = token
end for

close(INFILE)
printf(STDOUT, "Elapsed time: %f\n", time() - START)
