constant START = time()

-----------------------------------------------------
-- if a token begins with a -, is invalid
-- if a token ends with a -, is invalid
-- all tokens with no UCHAR, LCHAR or SQUOTE
-- are invalid
-----------------------------------------------------

-----------------------------------------------------
-- character-type list

constant LCHAR  = 1 -- lower-case character
constant UCHAR  = 2 -- upper-case character
constant NUMBER = 3 -- number
constant HYPHEN = 4 -- hyphen
constant SQUOTE = 5 -- single-quote

-- use for lower-to-upper conversion
constant UPPER = 'A'-'a'

sequence chars chars = repeat(0,255)

for n = 'a' to 'z' do
	chars[n] = LCHAR
end for

for n = 'A' to 'Z' do
	chars[n] = UCHAR
end for

for n = '0' to '9' do
	chars[n] = NUMBER
end for

chars['-'] = HYPHEN
chars['\''] = SQUOTE

-----------------------------------------------------
-- number of most frequent tokens to display

atom display  display  = 0

-----------------------------------------------------
-- command-line arguments

sequence cmd_line cmd_line = command_line()

if length(cmd_line) > 2 then
	cmd_line = cmd_line[3]
else
	cmd_line = ""
end if

-----------------------------------------------------
-- token sorting

integer over

sequence top20c -- top-20 count
sequence top20t -- top-20 words

top20c = repeat(0,20)
top20t = repeat({},20)

-----------------------------------------------------
-- tracking for characters while reading

integer cur_char  cur_char  = -1
integer prev_char prev_char = -1

sequence word word = repeat(0,20)

-----------------------------------------------------
-- count of tokens

integer total    total = 0  -- total number of tokens
integer unique   unique = 0 -- total unique tokens

sequence lengths lengths = repeat(0,20)

-----------------------------------------------------
-- token storage

sequence tokens tokens = repeat(repeat({},20),37)
sequence occurs occurs = repeat(repeat({},20),37)

-----------------------------------------------------
-- used to seek to the next valid token
-- seek to the next valid char in the file

procedure seek_token( integer f )
	while cur_char != -1 do
		if cur_char > 0 then
			if chars[cur_char] then
				exit
			end if
		end if
		cur_char = getc(f)
	end while
end procedure

-----------------------------------------------------
-- used to read through valid tokens
-- reads until no more valid chars are found

procedure skip_word( integer f )
	while cur_char != -1 do
		if cur_char > 0 then
			if chars[cur_char] = 0 then
				exit
			end if
		end if
		cur_char = getc(f)
	end while
end procedure

-----------------------------------------------------
-- retrieves an unknown token-word from the file
-- if 0 chars, or more than 20 chars, is invalid
-- if only numbers and dashes, is invalid
-- if quote occurs, is valid regardless

function get_word( integer f )

 integer i, valid
 integer ttype

	i = 1
	valid = 0

	-- collect chars until eof or invalid char
	while cur_char > 0 do

		ttype = chars[cur_char]

		-- check word-length
		if i = 21 and ttype != 0 then
			skip_word(f)
			i = 1 -- will return 0
			exit
		elsif ttype = LCHAR then
			word[i] = cur_char+UPPER
			i += 1
			valid = 1
		elsif ttype = UCHAR then
			word[i] = cur_char
			i += 1
			valid = 1
		elsif ttype = NUMBER then
			word[i] = cur_char
			i += 1
		elsif ttype = SQUOTE then
			valid = 1
		elsif ttype = HYPHEN then
			word[i] = cur_char
			i += 1
		else
			-- cur_char is no longer valid
			exit
		end if

		prev_char = cur_char
		cur_char = getc(f)

	end while

	-- check validity of token

	-- check prev_char for hyphen
	if prev_char = '-' then
		i = 1
	elsif not valid then
		return 0
	end if

	return i-1

end function

-----------------------------------------------------
-- store a token in the local tracking

procedure store_token( integer idx )

 integer char
 integer match

	char = word[1]

	if char >= '0' and char <= '9' then
		char -= '0'-1
	elsif char = '-' then
		char = 37
	else
		char -= 'A'-11
	end if

	match = find(word[1..idx],tokens[char][idx])
	if not match then
		tokens[char][idx] &= {word[1..idx]}
		occurs[char][idx] &= 1
	else
		occurs[char][idx][match] += 1
	end if

end procedure

-----------------------------------------------------
-- process the file, if valid

procedure scan_file( sequence fname )

 integer f
 integer i

	f = open(fname,"rb")
	if f != -1 then

		-- prime cur-char
		cur_char = getc(f)

		-- seek to the first valid token
		seek_token(f)

		while cur_char != -1 do

			if cur_char = '-' then

				-- word is invalid, regardless

				skip_word(f)

			else

				-- unknown if word is valid or not
				i = get_word(f)

				if i then
					lengths[i] += 1
					total += 1
					store_token(i)
				end if

			end if

			seek_token(f)

		end while

		close(f)
	end if

end procedure

-----------------------------------------------------
-----------------------------------------------------

-- process the file

scan_file( cmd_line )

-----------------------------------------------------
-- sort the words
-- count the unique words at the same time

for c = 1 to 37 do -- c = first-char of word

	for i = 1 to 20 do -- i = length of word

		for n = 1 to length(occurs[c][i]) do

			unique += 1

			over = find(1,occurs[c][i][n]>top20c)
			if over then

				top20c[over+1..20] = top20c[over..19]
				top20c[over] = occurs[c][i][n]

				top20t[over+1..20] = top20t[over..19]
				top20t[over] = tokens[c][i][n]

			end if

		end for

	end for

end for

-----------------------------------------------------
-- calculate the number of words to display

display = floor(total/1000)+1
if display > 20 then
	display = 20
end if


-- display the results

printf(1,"Jason Mirwald %s\n",{cmd_line})
printf(1,"Total: %7d, Unique: %7d\n",{total,unique})
puts(1,"\n")

-- display the most common unique words

for n = 1 to display do
	printf(1, "%02d %-20s %d\n" , {n,top20t[n],top20c[n]})
end for
puts(1,"\n")

-- display the frequency by length

for n = 1 to 20 do
	printf(1, "%02d %d\n", {n,lengths[n]})
end for
puts(1,"\n")

printf(1, "Elapsed time: %f\n", time() - START)
