------------------------------------------------------------------------------
--Parser version 4
------------------------------------------------------------------------------
--Not a state machine
--Handles NULL chars, and chars 128-255
--
------------------------------------------------------------------------------
without type_check
without warning
--<WORD PARSER Version 4>======================================================
--Global variables. Not wrapped in functions, to increase speed
--File handle. Remains non-zero as long as the input file is open
global integer fn
--Input file name. Set by open_file
global sequence file_name
-------------------------------------------------------------------------------
--Global constants
global constant MAX_CHAR = 127, MAX_WORD_SIZE = 20

--Internal constants
constant zero_array = repeat(0, MAX_CHAR)
-------------------------------------------------------------------------------
--File opening function
--See what the name of the file is. If none, abort.
global procedure open_file()
    sequence cl 
    cl = command_line()
    if length(cl) < 3 then 
        abort(0) 
    end if
    --Open the file for reading. If it fails, abort.
    file_name = cl[3]
    fn = open(file_name, "r")
    if fn = -1 then 
        abort(0) 
    end if
end procedure
-------------------------------------------------------------------------------
--Bit masks for all the characters, faster than manual comparison
--All the masks are MAX_CHAR in length, and elements are 1 or 0.
--Numbers
    global sequence number_char number_char = zero_array
    for c = '0' to '9' do number_char[c] = 1 end for
--Uppercase letters (lowercase are conditioned to become uppercase)
    global sequence letter_char letter_char = zero_array 
    for c = 'A' to 'Z' do letter_char[c] = 1 end for
    global sequence lower_letter_char lower_letter_char = zero_array
    for c = 'a' to 'z' do lower_letter_char[c] = 1 end for
--Dash
    global sequence dash_char dash_char = zero_array
    dash_char['-'] = 1
--Blanks - characters that can be between words and the inverse
    global sequence blank_char 
    blank_char = not ( number_char or letter_char or lower_letter_char or dash_char )
    global constant not_blank_char = not blank_char
--Number or a letter
    global constant number_or_letter = number_char or letter_char
-------------------------------------------------------------------------------
--Optimised 'function' to make all letters upper-case. (used within get_next_word)
--**HANDLES characters above 127 - converts them all to 127**
--The input must be offset by 1 before converting, to deal with null pointers.
sequence to_upper
to_upper = repeat(MAX_CHAR, 256)
for c = 1 to MAX_CHAR do    --Any characters above MAX_CHAR are left at MAX_CHAR, as is the NULL char
    if lower_letter_char[c] then --If a lower letter, replace with an upper
        to_upper[c+1] = c - ( 'a' - 'A' )
    else                          --Otherwise just copy it in normally
        to_upper[c+1] = c
    end if
end for


-------------------------------------------------------------------------------
--Internal variables used by get_next_word
--(code is faster if it doesn't have to keep allocating and deallocating
-- space for variables)
integer ch          --current character read from file
sequence buffer     --temporary buffer to store letters in
integer buffer_size --amount of the buffer being used (length of token)

--Internal flags
integer first_was_quote --flag for if the word had a leading quote
integer last_was_quote --flag for if the word had a trailing quote
integer has_letters --flag for if the word has some letters in it

--Initialise buffer only once
buffer      = repeat(0, MAX_WORD_SIZE)
-------------------------------------------------------------------------------
--get_next_word() 
--Returns the next valid word in the file. If no more valid words exist, 
--it closes the file and sets fn to zero as a flag bit.
--
--Once fn is closed, calling get_next_word() again will cause an error.
integer temp
global function get_next_word()
    --Initial values of internal variables and flags
    buffer_size = 0
    first_was_quote = 0
    last_was_quote = 0
    has_letters = 0

    ch = getc(fn)   --get the next character
    --Iterate through the file as long as more characters exist
    while ch != -1 do
        --Transform any lower-case chars to upper-case
        ch = to_upper[ch+1]

        --If the current char is a quote, remember but don't add it
        if ch = '\'' then
            --If at the beginning of the word:
            if not buffer_size then first_was_quote = 1 end if
            --Might be the end...
            last_was_quote = 1
            --Counts as a letter though...
            has_letters = 1
            
        --If the current char isn't blank, add it to the buffer
        elsif not_blank_char[ch] then
            --No trailing quote, this character comes after it.
            last_was_quote = 0
            --If the ch is a letter, remember
            if letter_char[ch] then has_letters = 1 end if
            --Add the character to the buffer				
            buffer_size += 1
            --If the word is too long, clear the buffer,
            --and cycle through until a blank is found
            if buffer_size > MAX_WORD_SIZE then 
                buffer_size = 0
                has_letters = 0
                first_was_quote = 0
                last_was_quote = 0
                while not_blank_char[ch] do 
                    ch = getc(fn)
                end while
            else buffer[buffer_size] = ch               --Add char to buffer
            end if
        --If a word is in the buffer and the next letter is a blank (implied), a word is found!
        --If neither the first nor last letters are dashes, and the word contains a letter or quote:
        elsif buffer_size and has_letters and
            (first_was_quote or buffer[1] != '-') and 
            (last_was_quote or buffer[buffer_size] != '-') then

            --Return the word
            return buffer[1..buffer_size]
        --If the word wasn't valid, continue searching
        else
            buffer_size = 0
            has_letters = 0
            first_was_quote = 0
            last_was_quote = 0
        end if

        --Get the next character from the file
        ch = getc(fn)
    end while
    
    --If the end of the file was reached in the middle of the word...
    --and neither the first nor last letters are dashes, and the word contains a letter or quote, it's valid!
    if buffer_size and has_letters and
            (first_was_quote or buffer[1] != '-') and 
            (last_was_quote or buffer[buffer_size] != '-') then
        return buffer[1..buffer_size]
    end if
    
    --The end of the file was reached, close it.
    close(fn)
    fn = 0
    return {} --Not a word
end function
--</WORD PARSER>===============================================================