constant START = time()
without warning
without type_check

constant cl = command_line()
if length(cl) = 2 then 
  printf(2, "usage: %s %s <filename>\n\n", cl)
  abort(1)
end if
constant 
  FN = cl[3],   -- file name
  MAX = 20,     -- maximum token length
  MAXRANK = 20, -- maximum number of ranked words
  IB = ('Z'+1), -- initial letter buckets
  ITLEN = (MAX+1)*IB, -- initial tree length
  LETTERS = 1,
  DIGITS_HYPHEN = 2,
  QUOTE = 3

printf(1, "Pete Eberlein %s\n", {FN})

integer Total_Count, Unique_Count, tstlen, tlen, len, toki
sequence tf, tok, tst, map, C, T
Total_Count = 0
Unique_Count = 0
tf = repeat(0, MAX)	-- token frequency counters
tok = repeat(0, MAX)    -- token buffer
tstlen = 4096		-- ternary search tree allocated length
tst = repeat(0, tstlen) -- ternary search tree
tlen = ITLEN		-- ternary search tree used length

map = repeat(0,255)	-- map of character domains
map['a'..'z'] = LETTERS
map['A'..'Z'] = LETTERS
map['0'..'9'] = DIGITS_HYPHEN
map['-'] = DIGITS_HYPHEN
map['\''] = QUOTE


procedure scan_file()
  integer f, c, d, tl, i, n, c1, c2, ni

  f = open(FN, "rb")	-- open file
  if f=-1 then
    printf(2, "Could not open file: %s\n", {FN})
    abort(1)
  end if
  c = getc(f)
  tl = 0
  while c != -1 do  -- loop until EOF

    if c >= '\'' then
      d = map[c]
      -- perform first character tests
      if d = LETTERS then  -- letters
        tl = 1
	tok[tl] = c
        c = getc(f)

      elsif d = DIGITS_HYPHEN then  -- digits or hyphens
        if c = '-' then -- tokens starting with hyphen are delimiters
          while c >= '\'' and map[c] do  -- read the rest of it
            c = getc(f)
          end while
        else
          -- tokens composed completely of digits and hyphens are delimiters
          while c > '\'' and map[c] = DIGITS_HYPHEN do
            tl += 1
            if tl > MAX then   -- token too long
              tl = 0
              while c >= '\'' and map[c] do  -- read the rest of it
                c = getc(f)
              end while
              exit
            end if
            tok[tl] = c
            c = getc(f)
          end while
          if c < '\'' or not map[c] then  -- if next char isn't a token char,
            tl = 0               -- then it must have been all digits/hyphens
          end if
        end if
      elsif d then -- QUOTE
        c = getc(f)
      else 
        c = 0
      end if

      while c >= '\'' do
        d = map[c]   -- get character's domain from map
        if d = LETTERS then  -- letter
	  tl += 1
	  if tl > MAX then  -- token too long
	    tl = 0
            while c >= '\'' and map[c] do  -- read the rest of it
	      c = getc(f)
            end while
            exit
          end if
	  tok[tl] = c
          c = getc(f)

        elsif d = DIGITS_HYPHEN then -- number or hyphen
	  tl += 1
          if tl > MAX then  -- token too long
	    tl = 0
            while c >= '\'' and map[c] do  -- read the rest of it
	      c = getc(f)
            end while
            exit
          end if

	  tok[tl] = c  -- put the char in the token buffer
          if c = '-' then  -- tokens ending with hypen are delimiters
            c = getc(f)
            if c < '\'' or not map[c] then  -- next char is delimiter
              tl = 0
              exit
            end if
          else  
            c = getc(f)            
          end if

        elsif c = '\'' then
          c = getc(f)

        else
          exit -- otherwise it's a delmiter char

        end if
      end while

      if tl then  -- token length not zero, so we got valid token

	-- ternary search tree algorithm
	-- three branches at each node: char equal, less, or greater than

	--- NOTE: and_bits(#DF,tok[n]) automatically converts to uppercase
  	-- starting node index is based on token length and first char
	ni = tl*IB + and_bits(#DF,tok[1])
	if tl > 1 then
	  n = 2  -- start on the 2nd char since the first is in the node index
  	  c2 = and_bits(#DF,tok[n])	-- read the second character
	  while 1 do
  	    i = tst[ni] -- does the next node have a pointer?
            if not i then -- no: the token is unique
	    
  	      while n <= tl do -- read the rest of the token

  	        if tlen+4 > tstlen then  -- will the tree be too small
  	          tst &= repeat(0,tstlen)  -- double its size
  	          tstlen += tstlen
  	        end if

  	        tst[tlen] = and_bits(#DF,tok[n])  -- write the token char to the node
  	        tst[ni] = tlen  -- current index points to new node
  	        ni = tlen+1     -- next index points to "equal" pointer
  	        tlen += 4	-- increase the tree by amount used by node
  	        n += 1

  	      end while
  	      --tst[ni] = 1 -- set initial count of this token
  	      exit	-- we're done
  	    end if
  	      
  	    c1 = tst[i]  -- get the char of this tree node
  	    if c1 < c2 then
  	      ni = i + 2 -- next index points to "less than" pointer
  	    elsif c1 > c2 then
  	      ni = i + 3 -- next index points to "greater than" pointer
            else
  	      ni = i + 1 -- next index points to "equal" pointer
  	      if n = tl then  -- at the end of the token?
  	        --tst[ni] += 1  -- increment its count
  	        exit	-- we're done
  	      end if
  	      n += 1
  	      c2 = and_bits(#DF,tok[n])  -- get the next token char
  	    end if
  	  end while
	end if
	tst[ni] += 1  -- increment its count
	tl = 0	-- reset the token length
      end if
    end if
    c = getc(f)  -- read char from file
  end while
  close(f) -- close the file
end procedure

-- recursively look for leaf nodes that have the highest count
procedure node(integer i)
  integer saved_toki, c
  saved_toki = toki  -- save token index
  while toki < len do
    c = tst[i+3]
    if c then
      node(c)  -- search less-than nodes, recursively
    end if

    c = tst[i+2]
    if c then
      node(c)  -- search greater-than nodes, recursively
    end if

    toki += 1
    tok[toki] = tst[i]  -- store character to token

    i = tst[i+1]  -- search equal nodes, tail-recursively
  end while
  toki = saved_toki  -- restore token index
  
-- got a full token
--printf(1,"%-20s %d\n", {tok[1..len],i})
  Unique_Count += 1  -- increment unique count
  tf[len] += i       -- increment token length frequency 

  if i >= C[MAXRANK] then  -- count greater thank lowest rank count?
    for Rank = 1 to MAXRANK do
      if i > C[Rank] or -- count greater, or equal and ascending token order
        (i = C[Rank] and compare(tok,T[Rank]) = -1) then
        -- insert the count into the ranking and update tokens
        C[Rank..MAXRANK] = i & C[Rank..MAXRANK-1]
        T[Rank..MAXRANK] = {tok[1..len]} & T[Rank..MAXRANK-1]
        exit
      end if
    end for
  end if
end procedure

procedure show_results ()
  integer X, c
  
  -- It's faster to count Uniques, Totals, and Frequencies during the 
  -- the rank search, even though we have to use the max ranks initially.
  T = repeat("", MAXRANK)
  C = repeat(0, MAXRANK)
  
  -- calculate the ranks
  toki = 1
  len = 0
  for tl = IB to MAX*IB by IB do
    len += 1
    tok = repeat(0,len)
    for c2 = '-' to IB-1 do
      c = and_bits(#DF,c2)
      tok[toki] = c
      c = tst[tl+c]
      if c then
        node(c)
      end if
    end for
  end for
  
  for i = 1 to MAX do
    Total_Count += tf[i]   -- increment total count
  end for

  -- show some statistics
  printf(1, "Total: %7d, Unique: %7d\n", {Total_Count, Unique_Count})
  X = floor(Unique_Count/1000)+1
  if (X > MAXRANK) then X = MAXRANK end if

  -- show the ranks
  for rank = 1 to X do
    T[rank] += 32 * (T[rank] < 32) -- convert non-uppercase chars
    printf(1, "%02d %-20s %d\n" , {rank, T[rank], C[rank]})
  end for

  -- show the length frequencies
  for n = MAX to 1 by -1 do
    if tf[n] then
      for i = 1 to n do
        printf(1,"%02d %d\n", {i, tf[i]})
      end for
      exit
    end if
  end for
end procedure

scan_file()
show_results()

printf(1, "Elapsed time: %f\n", time() - START)
