constant START = time()

-- Contest #1 answer - Ricardo M. Forno

without type_check

constant DELIMITER = 0
constant LETTER = 8
constant QUOTE = 4
constant MINUS = 2
constant DIGIT = 1
constant MAXALLOW = 20 -- Maximum allowable length
constant TABLEN = 65536 -- Hash tables length. Must be a power of 2
constant MASK = TABLEN - 1 -- To obtain remainder mod TABLEN
constant SMALLTABLEN = 32768 -- To increase locality of reference
constant MASK1 = SMALLTABLEN - 1 -- To obtain remainder mod SMALLTABLEN
constant BIGTABLEN = TABLEN + SMALLTABLEN + 256
constant INCREMENT = SMALLTABLEN + 1

integer Wordcount, Unique
sequence Lengths, Words, Counts 

function read_command()
    -- Read command line
    sequence cmd
    cmd = command_line()
    if length(cmd) < 3 then
	puts(2, "Error: this program needs a command-line argument\n")
	abort(1)
    end if
    return cmd[3] -- File name
end function

function open_file(sequence filename)
-- Open data file
    integer fn -- File number
    fn = open(filename, "rb")
    if fn = -1 then
	puts(2, "Error: cannot read file " & filename & " for reading\n")
	abort(2)
    end if
    return fn
end function

function fakesort(integer x, sequence indexes)
-- Especialized insertion sort for this program
-- x is the number of desired results
-- grade is a vector of indexes to unique tokens
    integer k, j, val, test
    sequence grade
    grade = repeat(0, x + 1)
    k = 0
    for i = 1 to Unique do
	val = Counts[indexes[i]]
	j = k
	while j do
	    test = val - Counts[grade[j]]
	    if test < 0 then
		exit
	    elsif test = 0 then
		if compare(Words[indexes[i]], Words[grade[j]]) > 0 then
		    exit
		end if
	    end if
	    j -= 1
	end while
	if j < x then
	    j += 1
	    grade[j + 1 .. k + 1] = grade[j .. k]
	    grade[j] = indexes[i]
	    k += k < x
	end if
    end for
    return grade
end function

procedure showresults()
    integer x, g, k
    sequence grade, indexes
    x = floor(Unique / 1000) + 1
    if x > MAXALLOW then
	x = MAXALLOW
    elsif Unique = 0 then
	x = 0
    end if
    indexes = repeat(0, Unique)
    printf(1, "Total: %7d, Unique: %7d\n", {Wordcount, Unique})
    k = 0
    for i = 1 to 256 do -- Letters -> one-letter words
	Words[i] = {Words[i]}
    end for
    for i = 1 to BIGTABLEN do
	if Words[i][1] then
	    k += 1
	    indexes[k] = i
	end if
    end for
    grade = fakesort(x, indexes)
    for i = 1 to x do
	g = grade[i]
	printf(1, "%02d %-20s %d\n" , {i, Words[g], Counts[g]})
    end for
    for i = 1 to MAXALLOW do
	if Lengths[i] then
	    printf(1, "%02d %d\n", {i, Lengths[i]})
	end if
    end for
end procedure

procedure wordprocess(integer fn)
-- Process file
    integer c, wlength, ctype, hash, h, char, test, lastctype
    sequence translated, chartype, word, hashconstants, w
    -- Create translation and character type tables
    translated = repeat(0, 256)
    chartype = repeat(DELIMITER, 256) -- All delimiters at start
    for i = 1 to 256 do
	translated[i] = i - 1
    end for
    for i = 'a' + 1 to 'z' + 1 do -- To uppercase
	translated[i] = i + 'A' - 'a' - 1
	chartype[i] = LETTER
    end for
    for i = 'A'+ 1 to 'Z' + 1 do
	chartype[i] = LETTER
    end for
    for i = '0' + 1 to '9' + 1 do
	chartype[i] = DIGIT
    end for
    chartype['\'' + 1] = QUOTE
    chartype['-' + 1] = MINUS

    -- Create hashing table
    hashconstants = repeat(0, 256)
    for i = 1 to 256 do
	hashconstants[i] = rand(TABLEN)
    end for

    word = repeat(0, MAXALLOW)
    Lengths = repeat(0, MAXALLOW) -- Create lengths table
    Words = repeat(repeat(0, MAXALLOW), BIGTABLEN) -- Create words hash table
    for i = 1 to 256 do -- For one-letter words
	Words[i] = 0
    end for
    Counts = repeat(1, BIGTABLEN) -- Initialized to 1 to avoid an assignment...
    Wordcount = 0
    Unique = 0
    wlength = 0
    ctype = 0
    c = 0
    lastctype = 0
    test = 0
    hash = 0
    h = 0
    
    c = getc(fn) + 1
    if c = 0 then -- Test EOF
	return
    end if

    while 1 do
	while 1 do -- Skip delimiters
	    if c then -- Test EOF
		ctype = chartype[c]
		if ctype then -- As DELIMITER = 0, this tests for not DELIMITER
		    exit
		end if
		c = getc(fn) + 1
	    else
		return
	    end if
	end while
	
	if ctype = MINUS then
	    c = getc(fn) + 1
	    if c then -- Test EOF
		ctype = chartype[c]
		while ctype do -- If not delimiter
		    c = getc(fn) + 1
		    if c then -- Test EOF
			ctype = chartype[c]
		    else
			exit
		    end if
		end while
	    else
		exit
	    end if
	else
	    -- Special case for 1st character to gain some speed...
	    lastctype = ctype 
	    test = ctype
	    if ctype = QUOTE then
		wlength = 0
		hash = 0
	    else
		wlength = 1
		char = translated[c]
		word[1] = char
		hash = hashconstants[char] -- Hashing
	    end if
	    c = getc(fn) + 1 -- Get next character
	    if c then -- Test EOF
		ctype = chartype[c] -- Determine its type
	    else
		ctype = DELIMITER
	    end if
	    -- Characters after 1st
	    while ctype do -- Non delimiter characters
		lastctype = ctype
		test = or_bits(test, ctype)
		if ctype != QUOTE then
		    if wlength < MAXALLOW then
			wlength += 1
			char = translated[c]
			word[wlength] = char
			hash += hashconstants[char] -- Hashing
		    else
			test = 0
		    end if
		end if
		c = getc(fn) + 1 -- Get next character
		if c then -- Test EOF
		    ctype = chartype[c] -- Determine its type
		else
		    exit
		end if
	    end while
	    -- Now, we have a candidate token
	    if test > 4 then -- A complex test:
	    -- if the maximum allowed size of a token has been surpased, then
	    --  test = 0
	    -- if only digits appeared, then
	    --  test = 1
	    -- if only minus signs appeared, then
	    --  test = 2
	    -- if only minus signs and digits appeared, then
	    --  test = 3
	    -- if only quotes appeared, then
	    --  test = 4
	    -- else
	    --  test > 4
		if lastctype != MINUS then -- If last char not '-'
		    -- We got a token
		    Wordcount += 1 -- Count token
		    Lengths[wlength] += 1
		    if wlength = 1 then -- One-letter words
			char = word[1]
			if Words[char] then -- Letter already found
			    Counts[char] += 1
			else
			    Words[char] = char
			    Unique += 1
			end if
		    else
			h = and_bits(hash, MASK1) + 257
			w = Words[h]
			if w[1] then -- If bucket not empty
			    if equal(w, word) then -- If not a synonym
				Counts[h] += 1 -- Count unique words
			    else
				-- Process collisions
				hash = and_bits(hash, MASK) + INCREMENT
				h = hash
				while 1 do
				    w = Words[h]
				    if w[1] then -- If bucket not empty
					if equal(w, word) then -- Synonym?
					    Counts[h] += 1 -- Count unique
					    exit
					elsif h >= BIGTABLEN then -- Overflow?
					    h = INCREMENT
					else
					    h += 1
					    if h = hash then -- No room?
						puts(2, "Hash table full\n")
						abort(3)
					    end if
					end if
				    else -- New token
					Unique += 1
					-- Counts initialized to 1 at start
					Words[h] = word
					exit
				    end if
				end while
			    end if
			else -- New token
			    Unique += 1
			    -- Counts initialized to 1 at start
			    Words[h] = word
			end if
		    end if
		end if
		word[2 .. wlength] = 0 -- Sure wlength > 0
	    else
		word[1 .. wlength] = 0 -- Reset used locations
	    end if
	end if
    end while
end procedure

procedure contest1()
    sequence filename
    integer fn
    filename = read_command() -- Get file name
    fn = open_file(filename)
    
    -- Show my identity and the file name
    puts(1, "Ricardo M. Forno " & filename & "\n")
    wordprocess(fn)
    showresults()
    close(fn)
end procedure

-- Execute main procedure
contest1()

printf(1, "Elapsed time: %f\n", time() - START)
